Package: sfnetworks
Title: Tidy Geospatial Networks
Version: 0.6.6
Authors@R: 
    c(person(given = "Lucas",
             family = "van der Meer",
             role = c("aut", "cre"),
             email = "luukvandermeer@live.nl",
             comment = c(ORCID = "0000-0001-6336-8628")),
      person(given = "Lorena",
             family = "Abad",
             role = "aut",
             email = "lore.abad6@gmail.com",
             comment = c(ORCID = "0000-0003-0554-734X")),
      person(given = "Andrea",
             family = "Gilardi",
             role = "aut",
             email = "andrea.gilardi@unimib.it",
             comment = c(ORCID = "0000-0002-9424-7439")),
      person(given = "Robin",
             family = "Lovelace",
             role = "aut",
             email = "r.lovelace@leeds.ac.uk",
             comment = c(ORCID = "0000-0001-5679-6536")))
Maintainer: Lucas van der Meer <luukvandermeer@live.nl>
Description: Provides a tidy approach to spatial network
    analysis, in the form of classes and functions that enable a seamless
    interaction between the network analysis package 'tidygraph' and the
    spatial analysis package 'sf'.
License: Apache License (>= 2)
URL: https://luukvdmeer.github.io/sfnetworks/,
        https://github.com/luukvdmeer/sfnetworks
BugReports: https://github.com/luukvdmeer/sfnetworks/issues/
Depends: R (>= 3.6)
Imports: crayon, dplyr, graphics, igraph, lwgeom, rlang, sf, sfheaders,
        tibble, tidygraph, units, utils
Suggests: dbscan, fansi, ggplot2 (>= 3.0.0), knitr, purrr, rmarkdown,
        s2 (>= 1.0.1), spatstat.geom, spatstat.linnet, testthat, TSP
VignetteBuilder: knitr
ByteCompile: true
Encoding: UTF-8
LazyData: true
Config/roxygen2/version: 8.0.0
NeedsCompilation: no
Packaged: 2026-05-14 17:10:33 UTC; rstudio
Author: Lucas van der Meer [aut, cre] (ORCID:
    <https://orcid.org/0000-0001-6336-8628>),
  Lorena Abad [aut] (ORCID: <https://orcid.org/0000-0003-0554-734X>),
  Andrea Gilardi [aut] (ORCID: <https://orcid.org/0000-0002-9424-7439>),
  Robin Lovelace [aut] (ORCID: <https://orcid.org/0000-0001-5679-6536>)
Repository: CRAN
Date/Publication: 2026-05-14 21:40:02 UTC
Built: R 4.6.0; ; 2026-05-15 00:06:13 UTC; unix
