# Generated by roxygen2: do not edit by hand

S3method("st_agr<-",sfnetwork)
S3method("st_crs<-",sfnetwork)
S3method("st_geometry<-",sfnetwork)
S3method(as_sfnetwork,default)
S3method(as_sfnetwork,linnet)
S3method(as_sfnetwork,psp)
S3method(as_sfnetwork,sf)
S3method(as_sfnetwork,sfNetwork)
S3method(as_sfnetwork,sfc)
S3method(as_sfnetwork,sfnetwork)
S3method(as_sfnetwork,tbl_graph)
S3method(as_tbl_graph,sfnetwork)
S3method(as_tibble,sfnetwork)
S3method(morph,sfnetwork)
S3method(plot,sfnetwork)
S3method(print,morphed_sfnetwork)
S3method(print,sfnetwork)
S3method(st_agr,sfnetwork)
S3method(st_area,sfnetwork)
S3method(st_as_s2,sfnetwork)
S3method(st_as_sf,sfnetwork)
S3method(st_bbox,sfnetwork)
S3method(st_coordinates,sfnetwork)
S3method(st_crop,morphed_sfnetwork)
S3method(st_crop,sfnetwork)
S3method(st_crs,sfnetwork)
S3method(st_difference,morphed_sfnetwork)
S3method(st_difference,sfnetwork)
S3method(st_drop_geometry,sfnetwork)
S3method(st_filter,morphed_sfnetwork)
S3method(st_filter,sfnetwork)
S3method(st_geometry,sfnetwork)
S3method(st_intersection,morphed_sfnetwork)
S3method(st_intersection,sfnetwork)
S3method(st_intersects,sfnetwork)
S3method(st_is,sfnetwork)
S3method(st_is_valid,sfnetwork)
S3method(st_join,morphed_sfnetwork)
S3method(st_join,sfnetwork)
S3method(st_m_range,sfnetwork)
S3method(st_nearest_points,sfnetwork)
S3method(st_network_bbox,sfnetwork)
S3method(st_network_blend,sfnetwork)
S3method(st_network_cost,sfnetwork)
S3method(st_network_join,sfnetwork)
S3method(st_network_paths,sfnetwork)
S3method(st_normalize,sfnetwork)
S3method(st_precision,sfnetwork)
S3method(st_reverse,sfnetwork)
S3method(st_sample,sfnetwork)
S3method(st_set_precision,sfnetwork)
S3method(st_shift_longitude,sfnetwork)
S3method(st_simplify,sfnetwork)
S3method(st_transform,sfnetwork)
S3method(st_wrap_dateline,sfnetwork)
S3method(st_z_range,sfnetwork)
S3method(st_zm,sfnetwork)
S3method(unmorph,morphed_sfnetwork)
export("%>%")
export(activate)
export(active)
export(as_sfnetwork)
export(edge_azimuth)
export(edge_circuity)
export(edge_contains)
export(edge_contains_properly)
export(edge_covers)
export(edge_crosses)
export(edge_displacement)
export(edge_equals)
export(edge_intersects)
export(edge_is_covered_by)
export(edge_is_disjoint)
export(edge_is_within)
export(edge_is_within_distance)
export(edge_length)
export(edge_overlaps)
export(edge_touches)
export(is.sfnetwork)
export(node_M)
export(node_X)
export(node_Y)
export(node_Z)
export(node_equals)
export(node_intersects)
export(node_is_covered_by)
export(node_is_disjoint)
export(node_is_within)
export(node_is_within_distance)
export(node_touches)
export(sf_attr)
export(sfnetwork)
export(st_network_bbox)
export(st_network_blend)
export(st_network_cost)
export(st_network_join)
export(st_network_paths)
export(to_spatial_contracted)
export(to_spatial_directed)
export(to_spatial_explicit)
export(to_spatial_neighborhood)
export(to_spatial_shortest_paths)
export(to_spatial_simple)
export(to_spatial_smooth)
export(to_spatial_subdivision)
export(to_spatial_subset)
export(to_spatial_transformed)
importFrom(crayon,silver)
importFrom(dplyr,across)
importFrom(dplyr,bind_rows)
importFrom(dplyr,full_join)
importFrom(dplyr,group_by)
importFrom(dplyr,group_indices)
importFrom(dplyr,group_size)
importFrom(dplyr,group_split)
importFrom(dplyr,mutate)
importFrom(graphics,plot)
importFrom(igraph,"edge_attr<-")
importFrom(igraph,"graph_attr<-")
importFrom(igraph,"vertex_attr<-")
importFrom(igraph,E)
importFrom(igraph,V)
importFrom(igraph,adjacent_vertices)
importFrom(igraph,all_shortest_paths)
importFrom(igraph,all_simple_paths)
importFrom(igraph,contract)
importFrom(igraph,count_components)
importFrom(igraph,decompose)
importFrom(igraph,degree)
importFrom(igraph,delete_edge_attr)
importFrom(igraph,delete_edges)
importFrom(igraph,delete_vertex_attr)
importFrom(igraph,delete_vertices)
importFrom(igraph,distances)
importFrom(igraph,ecount)
importFrom(igraph,edge.attributes)
importFrom(igraph,edge_attr)
importFrom(igraph,edge_attr_names)
importFrom(igraph,ends)
importFrom(igraph,gorder)
importFrom(igraph,graph_attr)
importFrom(igraph,gsize)
importFrom(igraph,igraph_opt)
importFrom(igraph,igraph_options)
importFrom(igraph,incident_edges)
importFrom(igraph,induced_subgraph)
importFrom(igraph,is_bipartite)
importFrom(igraph,is_connected)
importFrom(igraph,is_dag)
importFrom(igraph,is_directed)
importFrom(igraph,is_simple)
importFrom(igraph,make_graph)
importFrom(igraph,shortest_paths)
importFrom(igraph,simplify)
importFrom(igraph,vcount)
importFrom(igraph,vertex_attr)
importFrom(igraph,vertex_attr_names)
importFrom(igraph,which_loop)
importFrom(igraph,which_multiple)
importFrom(lwgeom,st_geod_azimuth)
importFrom(rlang,"!!")
importFrom(rlang,":=")
importFrom(sf,"st_agr<-")
importFrom(sf,"st_crs<-")
importFrom(sf,"st_geometry<-")
importFrom(sf,"st_precision<-")
importFrom(sf,sf_use_s2)
importFrom(sf,st_agr)
importFrom(sf,st_area)
importFrom(sf,st_as_s2)
importFrom(sf,st_as_sf)
importFrom(sf,st_as_sfc)
importFrom(sf,st_bbox)
importFrom(sf,st_cast)
importFrom(sf,st_centroid)
importFrom(sf,st_collection_extract)
importFrom(sf,st_combine)
importFrom(sf,st_contains)
importFrom(sf,st_contains_properly)
importFrom(sf,st_coordinates)
importFrom(sf,st_covered_by)
importFrom(sf,st_covers)
importFrom(sf,st_crop)
importFrom(sf,st_crosses)
importFrom(sf,st_crs)
importFrom(sf,st_difference)
importFrom(sf,st_disjoint)
importFrom(sf,st_distance)
importFrom(sf,st_drop_geometry)
importFrom(sf,st_equals)
importFrom(sf,st_filter)
importFrom(sf,st_geometry)
importFrom(sf,st_intersection)
importFrom(sf,st_intersects)
importFrom(sf,st_is)
importFrom(sf,st_is_longlat)
importFrom(sf,st_is_valid)
importFrom(sf,st_is_within_distance)
importFrom(sf,st_join)
importFrom(sf,st_length)
importFrom(sf,st_line_merge)
importFrom(sf,st_m_range)
importFrom(sf,st_nearest_feature)
importFrom(sf,st_nearest_points)
importFrom(sf,st_normalize)
importFrom(sf,st_overlaps)
importFrom(sf,st_precision)
importFrom(sf,st_reverse)
importFrom(sf,st_sample)
importFrom(sf,st_set_precision)
importFrom(sf,st_sf)
importFrom(sf,st_sfc)
importFrom(sf,st_shift_longitude)
importFrom(sf,st_simplify)
importFrom(sf,st_touches)
importFrom(sf,st_transform)
importFrom(sf,st_within)
importFrom(sf,st_wrap_dateline)
importFrom(sf,st_z_range)
importFrom(sf,st_zm)
importFrom(sfheaders,sf_to_df)
importFrom(sfheaders,sfc_linestring)
importFrom(sfheaders,sfc_point)
importFrom(sfheaders,sfc_to_df)
importFrom(stats,median)
importFrom(tibble,as_tibble)
importFrom(tidygraph,"%>%")
importFrom(tidygraph,.G)
importFrom(tidygraph,.graph_context)
importFrom(tidygraph,activate)
importFrom(tidygraph,active)
importFrom(tidygraph,as_tbl_graph)
importFrom(tidygraph,graph_join)
importFrom(tidygraph,morph)
importFrom(tidygraph,mutate)
importFrom(tidygraph,node_distance_from)
importFrom(tidygraph,node_distance_to)
importFrom(tidygraph,reroute)
importFrom(tidygraph,tbl_graph)
importFrom(tidygraph,unmorph)
importFrom(tidygraph,with_graph)
importFrom(units,as_units)
importFrom(units,deparse_unit)
importFrom(units,drop_units)
importFrom(units,set_units)
importFrom(utils,capture.output)
importFrom(utils,head)
importFrom(utils,packageVersion)
importFrom(utils,tail)
